/*
    Copyright 2004,2005,2006 Luigi Auriemma

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA

    http://www.gnu.org/licenses/gpl.txt
*/


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include "crc32.h"

#ifdef WIN32
    #include <winsock.h>
    #include <malloc.h>
    #include <direct.h>
    #include "winerr.h"

    #define close       closesocket
    #define sleep   Sleep
    #define strncasecmp strnicmp
    #define strcasecmp  stricmp
    #define strcasestr  stristr
#else
    #include <unistd.h>
    #include <sys/socket.h>
    #include <arpa/inet.h>
    #include <netdb.h>
#endif



#define VER         "0.4.1"
#define BUFFSZ      8192
#define FILESZ      32768
#define PK1OFF      29
#define PK1         0x04034b50      /* local_head */
#define PK2         0x02014b50      /* central_dir */
#define PK3         0x06054b50      /* end_central_dir */
#define PK4         0x05054b50      /* digital sign */
#define PORT        80
#define PPORT       8080
//#define OK206       " 206"
#define OK206       " 20"
#define CDIRSZ      sizeof(struct central_dir)
#define ECDIRSZ     sizeof(struct end_central_dir)
#define TOTCDIRSZ   (CDIRSZ + central_dir->name_len + central_dir->extra_len + central_dir->comm_len)
#define MAXNAMESZ   1024            /* operating systems support max 512 as path/file size */
#define DUMPERR     "\n##### DUMP #####\n%s\n##### DUMP #####\n\n"
#define INVALIDZIP  "\nError: this is not a valid ZIP file\n"
#define NORESUME    "\nError: Server doesn't support resume. Is impossible to continue\n"
#define UNKNERR     "\nError: Unknown problem during the download of the file\n" \
                    "(no data, probably not resuming support... you could try with -k)\n"
#define CONTLEN     "Content-Length: "
#define CONTRANGE   "Content-Range: bytes 0-0/"
#define GET         "GET /%s HTTP/1.0\r\n" \
                    "Host: %s\r\n" \
                    "Range: bytes=%u-%u\r\n" \
                    "Connection: Keep-Alive\r\n" \
                    "\r\n"
#define GETPROXY    "GET http://%s:%hu/%s HTTP/1.0\r\n" \
                    "Host: %s\r\n" \
                    "Range: bytes=%u-%u\r\n" \
                    "Connection: Keep-Alive\r\n" \
                    "\r\n"
#define startsck(x) sd = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP); \
                    if(sd < 0) std_err(); \
                    fputs("Connecting... ", stdout); \
                    err = connect(sd, (struct sockaddr *)&peer, sizeof(peer)); \
                    if(err < 0) std_err(); \
                    fputs("OK\n", stdout);
#define initsock(x) if(!keepalive) { \
                        if(sd) { \
                            close(sd); \
                        } \
                        startsck(x); \
                    } \
                    err = send(sd, buff, len, 0); \
                    if(err < 0) std_err();



u_char *read_http(int sd, u_int size);
u_short http2file(u_char *tempname, u_short namesz, u_int offset, u_int size, u_char *central, u_char extraredo);
int show_zip(u_char *central);
u_int calc_crc(u_char *filename);
u_char *spc_conv(u_char *arg);
void geturl(char *buff);
u_int resolv(char *hostchar);
char *maketime(u_int timedate);
void io_err(void);
void std_err(void);



#pragma pack(2)



struct central_dir {
    u_int  sign;
    u_short ver_made;
    u_short ver_ext;
    u_short flag;
    u_short method;
    u_int  timedate;
    u_int  crc;
    u_int  comp_size;
    u_int  uncomp_size;
    u_short name_len;
    u_short extra_len;
    u_short comm_len;
    u_short disk_start;
    u_short int_attr;
    u_int  ext_attr;
    u_int  offset;
    /* filename */
    /* extra */
    /* comment */
} *central_dir;



struct end_central_dir {
    u_short disk_num;
    u_short disk_num_start;
    u_short tot_ent_disk;
    u_short tot_ent_dir;
    u_int  dir_size;
    u_int  offset;
    u_short comm_len;
    /* comment */
} *end_central_dir;



char    *host,
        *connection,
        *uri;
int     port = PORT,
        proxy = 0,
        lamepatch = 1,
        portconn,
        percentage = 1,
        overwritefile = 0,
        filecompare = 0,
        extracheck = 0;
u_int  totdl = 0;
u_char  *dlarg = 0;
struct  sockaddr_in peer;



int main(int argc, char *argv[]) {
    u_char  *buff,
            *ptr = 0,
            *movederr = 0,
            *newarg,
            *proxyarg = 0,
            *end_dir_ptr,
            *workdir = 0;
    u_int  csz,
            esz;
    int     sd,
            err,
            len,
            contsz,
            keepalive = 1,
            sizeonly = 0,
            verboseerr = 0,
            method2 = 0,
            i,
            pck3off;

#ifdef WIN32
    WSADATA    wsadata;
    WSAStartup(MAKEWORD(1,0), &wsadata);
#endif


    setbuf(stdout, NULL);

    fputs("\n"
        "Zipweb "VER"\n"
        "by Luigi Auriemma\n"
        "e-mail: aluigi@autistici.org\n"
        "web:    aluigi.org\n"
        "\n", stdout);

    if(argc < 2) {
        printf("\nUsage: %s [options] <URL>\n"
            "\n"
            "Options:\n"
            "-p PROXY[:PORT] proxy support (default port %d)\n"
            "-w DIRECTORY    working directory for download and file comparison (current)\n"
            "-d FILENAME     download a file! Case insensitive and partial name comparison,\n"
            "                that means you can specify simply a part of the filename (cool)\n"
            "-i              interactive download: you can choose what files to download\n"
            "-s              shows only the filesize and exits (very very useful and can be\n"
            "                used with any file)\n"
            "-o              overwrites the existent files (if you choose to download a\n"
            "                file) without asking confirmation\n"
            "-f              file compare: will do a checksum comparison between the files\n"
            "                stored in the remote zip and your local files on the disk\n"
            "-v              verbose error messages\n"
            "\n"
            "Compatibility options (rare):\n"
            "-e              if you download a file that cannot be unzipped use this option,\n"
            "                will be made a check and will be redownloaded if wrong (useful)\n"
            "-l              doesn't apply the automatic fix for extra data equal to 0xd\n"
            "-c              doesn't show the file download percentage (less CPU?)\n"
            "-k              keep-alive disactivation. Useless because it is automatic\n"
            "-2              method 2, rarely useful for some rare bad zip files\n"
            "\n"
            "Note: use (\") around the URL if there are spaces or strange chars in it\n"
            "\n",
            argv[0], PPORT);
        exit(1);
    }

    argc--;
    for(i = 1; i < argc; i++) {
        switch(argv[i][1]) {
            case 'p': proxyarg = argv[++i]; break;
            case 'w': workdir = argv[++i]; break;
            case 'd': dlarg = argv[++i]; break;
            case 'i': dlarg = ""; break;
            case 's': sizeonly = 1; break;
            case 'o': overwritefile = 1; break;
            case 'f': filecompare = 1; break;
            case 'v': verboseerr = 1; break;
            case 'e': extracheck = 1; break;
            case 'l': lamepatch = 0; break;
            case 'c': percentage = 0; break;
            case 'k': keepalive = 0; break;
            case '2': method2 = 1; break;
            default: {
                printf("\nError: wrong argument (%s)\n", argv[i]);
                exit(1);
            }
        }
    }

    newarg = spc_conv(argv[argc]);    /* spaces to %20 */
    geturl(newarg);

        /* proxy */
    if(proxyarg) {
        portconn = PPORT;
        connection = strchr(proxyarg, ':');
        if(connection) {
            *connection = 0;
            portconn = atoi(connection + 1);
        }
        connection = proxyarg;
        proxy = 1;
    } else {
        connection = host;
        portconn = port;
    }

    fputs("Resolving hostname...\n", stdout);
    peer.sin_addr.s_addr = resolv(connection);
    peer.sin_port        = htons(portconn);
    peer.sin_family      = AF_INET;

        /* start socket */
    startsck(sd);

    printf("\n"
        "Connected to %s:%hu\n"
        "Host: %s:%hu\n"
        "URI:  /%s\n\n",
        inet_ntoa(peer.sin_addr),
        portconn,
        host,
        port,
        uri);

    if(workdir) {
        printf("\nChanging working directory: %s... ", workdir);
        err = chdir(workdir);
        if(err < 0) std_err();
        fputs("ok\n", stdout);
    }

    buff = malloc(BUFFSZ + 1);
    if(!buff) std_err();

        /* ------------------------------- */
        /* get filesize and resume support */
        /* ------------------------------- */

    if(!proxy) {
        len = snprintf(buff, BUFFSZ, GET, uri, host, 0, 0);
    } else {
        len = snprintf(buff, BUFFSZ, GETPROXY, host, port, uri, host, 0, 0);
    }
    err = send(sd, buff, len, 0);
    if(err < 0) std_err();

    len = 0;
    while(1) {
        err = recv(sd, buff + len, BUFFSZ - len, 0);
        if(err < 0) std_err();
        if(!err) {
            if(keepalive && !len) {
                fputs("\nError: strange connection error, I try to disable keep-alive and reconnect\n", stdout);
                if(verboseerr) printf(DUMPERR, buff);
                keepalive = 0;
                close(sd);
                startsck(sd);
                err = send(sd, buff, len, 0);
                if(err < 0) std_err();
                len = 0;
                continue;
            } else if(!len) {
                *buff = 0;   // no data
                break;
            } else break;
        }
        len += err;
        buff[len] = 0;

        ptr = strstr(buff, "\r\n\r\n");
        if(ptr) break;
    }

    if(keepalive) fputs("- Info: Keep-Alive is active\n", stdout);
        else fputs("- Info: Keep-Alive is not active\n", stdout);

    if(ptr) {
        ptr += 4;
        if((*ptr == '<') || (buff[9] != '2')) {
                /* yeah I manage also the moving errors (301, 302 and so on)... cool */
            if(buff[9] == '3') {
                movederr = strcasestr(buff, "location: ");
                if(movederr) {
                    movederr += 10;
                    ptr = strchr(movederr, '\n');
                    if(ptr) {
                        err = *ptr;     // I must restore it later
                        *ptr = 0;
                    }
                } else {
                    movederr = strchr(ptr, '\"');
                    if(movederr) {
                        movederr++;
                        ptr = strchr(movederr, '\"');
                        if(ptr) {
                            err = *ptr; // I must restore it later
                            *ptr = 0;
                        }
                    } else movederr = "unknown";
                }

                printf("\n"
                    "Error: seems the file has been moved to:\n"
                    "\n  %s\n\n"
                    "(Use the -v flag if you need more details)\n",
                    movederr);
                *ptr = err;
            } else {
                printf("\nError: File doesn't seem to exist on the server (error %.3s)\n",
                    buff + 9);
            }
            if(verboseerr) printf(DUMPERR, buff);
            goto esci;
        }
    }

        /* check resume support or file presence */
        /* in a perfect world we should check 206 but not all the webservers */
        /* are the sames so we check 20 (as 200 and 206) */
    if(memcmp(buff + 8, OK206, sizeof(OK206) - 1)) {
        ptr = strcasestr(buff, CONTLEN);
        if(ptr) {
            if(atol(ptr + (sizeof(CONTLEN) - 1)) > 2) {
                fputs(NORESUME, stdout);
                if(verboseerr) printf(DUMPERR, buff);
                goto esci;
            }
        }
        printf("\nError: Server or unknown error, dump follows\n"
            DUMPERR, buff);
        goto esci;
    }

        /* check filesize and resume support */
    ptr = strcasestr(buff, CONTRANGE);
    if(ptr) {
        contsz = atol(ptr + (sizeof(CONTRANGE) - 1));
        printf("\n\tTotal filesize: %u\n\n", contsz);
    } else {
        fputs(NORESUME, stdout);
        if(verboseerr) printf(DUMPERR, buff);
        goto esci;
    }

    if(sizeonly) return(0);

        /* ----------------------------------------- */
        /* downloading of the latest part of the zip */
        /* ----------------------------------------- */

    esz = (0xffff + ECDIRSZ);   /* max end_central_dir size and also a bit of space for HTTP headers */
    if(esz > contsz) {
        csz = 0;
        esz = contsz;
    } else csz = contsz - esz;

        /* csz is the total filesize less the part we are downloading */

    if(!proxy) {
        len = snprintf(buff, BUFFSZ, GET, uri, host, csz, contsz);
    } else {
        len = snprintf(buff, BUFFSZ, GETPROXY, host, port, uri, host, csz, contsz);
    }

    initsock(sd);
    ptr = read_http(sd, esz);
    if(!ptr) {
        fputs("\n- Server doesn't support Keep-Alive, so I disable it\n", stdout);
        keepalive = 0;
        initsock(sd);
        ptr = read_http(sd, esz);
        if(!ptr) {
            fputs(UNKNERR, stdout);
            if(verboseerr) fputs("No verbose error here\n", stdout);
            goto esci;
        }
    }

        /* search of 0x06054b50, end_central_dir */

    end_dir_ptr = 0;
    ptr += (esz - 4);
    for(pck3off = esz - 4; pck3off > 0; pck3off--, ptr--) {
        if(*(u_int *)ptr == PK3) {
            end_dir_ptr = ptr;
            break;
        }
    }
    ptr -= pck3off;

    if(!end_dir_ptr) {
        fputs("\n"
            "Error: remote file doesn't seem to be a real zip file\n"
            "       Is also possible that the file doesn't exist on the server\n", stdout);
        if(verboseerr) {
            ptr[esz] = 0;
            printf(DUMPERR, ptr);
        }
        goto esci;
    }

    end_central_dir = (struct end_central_dir *)(end_dir_ptr + 4);

    printf("\n"
        "- ZIP central directory:\n"
        "  Size: %-10u  Offset: %-10u (0x%08x)\n",
        end_central_dir->dir_size,
        end_central_dir->offset,
        end_central_dir->offset);

    if(end_central_dir->comm_len) {
        ptr[esz] = 0;        // for possible comment
        printf("\n"
            "-=======================-C-O-M-M-E-N-T-========================-\n"
            "%s\n"
            "-=======================---------------========================-\n",
            end_dir_ptr + ECDIRSZ + 4);
    }

        /* that's the method 2, probably useless */
    if(method2) {
        i = contsz - end_central_dir->offset;
    } else {
        i = end_central_dir->dir_size;
        end_central_dir->offset = csz - i + pck3off;
    }

        /* if we have found PK2 */
    if(i <= esz) {
        close(sd);
        err = show_zip(ptr + esz - i);
        if(err < 0) {
            ptr = 0;
            for(; pck3off > 0; pck3off--, end_dir_ptr--) {
                if(*(u_int *)end_dir_ptr == PK2) ptr = end_dir_ptr;
            }
            if(!ptr) {
                fputs(INVALIDZIP, stdout);
                exit(1);
            }
            err = show_zip(ptr);
            if(err < 0) {
                fputs(INVALIDZIP, stdout);
                exit(1);
            }
        }
    } else {
        printf("\n- Downloading of the remaining zip informations from offset: %08x\n",
            end_central_dir->offset);

        if(!proxy) {
            len = snprintf(buff, BUFFSZ, GET, uri, host, end_central_dir->offset, contsz);
        } else {
            len = snprintf(buff, BUFFSZ, GETPROXY, host, port, uri, host, end_central_dir->offset, contsz);
        }

        initsock(sd);

            /* yes I know that the previous buffer has not been unallocated... */
        ptr = read_http(sd, i);
        if(!ptr) {
            fputs(UNKNERR, stdout);
            if(verboseerr) fputs("No verbose error here\n", stdout);
            goto esci;
        }
        close(sd);
        err = show_zip(ptr);
        if(err < 0) {
            fputs(INVALIDZIP, stdout);
            exit(1);
        }
    }

    if(dlarg) printf("\nDownloaded %u files\n\n", totdl);
        else fputc('\n', stdout);

    return(0);


/* goto esci */
esci: {
    if(sd) close(sd);
    return(1);
}

}



u_char *read_http(int sd, u_int size) {
    u_int          len,
                    head = 0,
                    sizecp = size,  // sizecp is the total size of the buffer
                    tot;
    static  u_char  *buff = 0;
    u_char          *ptr = 0;
    int             err;

    buff = malloc(size + 1);
    if(!buff) std_err();

    printf("Downloading %u bytes", size);

    len = 0;
    tot = 0;
    while(tot < size) {
        err = recv(sd, buff + len, sizecp - len, 0);
        if(err < 0) std_err();
        if(!err) break;
        len += err;

        if(!head) {
            buff[len] = 0;
            ptr = strstr(buff, "\r\n\r\n");
            if(ptr) {
                ptr += 4;
                tot = len - (ptr - buff);
                sizecp += (ptr - buff);
                buff = realloc(buff, sizecp + 1);
                if(!buff) std_err();
                head = 1;
            }
        } else {
            tot += err;
            fputc('.', stdout);
        }
    }

    if(tot < size) {
        free(buff);
        return(NULL);
    }

    return(ptr);
}



u_short http2file(u_char *tempname, u_short namesz, u_int offset, u_int size, u_char *central, u_char extraredo) {
    int     sd = 0,
            err,
            len;
    u_int  length,
            length2,
            realfile,
            currsz,
            fsize,
            tmp;
    u_char  *buff,
            *filename,
            *stri,
            *contlen;
    u_short tmptest;
    FILE    *fd = 0;
    struct  stat    xstat;

    if(!*dlarg) {
        fputs("- Interactive: Do you want to download it? (y/n)\n", stdout);
        fflush(stdin);
        do { err = fgetc(stdin); } while(err == '\n');
        if((err != 'y') && (err != 'Y')) {
            fputs("  *SKIPPED*\n", stdout);
            return(0);
        } else fputs("  *DOWNLOADING*\n", stdout);
    }

    filename = alloca(namesz + 6);
    if(!filename) std_err();

    memcpy(filename, tempname, namesz);
    stri = filename;
    for(err = 0; err < namesz; err++, stri++) {
        if(*stri == '\\') *stri = '.';
        if(*stri == '/') *stri = '.';
    }
    memcpy(stri, ".zip\0", 5);

    printf("\n"
        "***************************************************************\n"
        "Download file as: %s\n",
        filename);

    err = stat(filename, &xstat);
    if(err >= 0) {
        if(!overwritefile && !extraredo) {
            if(xstat.st_size == (size + TOTCDIRSZ + ECDIRSZ + 5)) {
                fputs("\n"
                    "Info: a file with the same NAME *and* SIZE already exists\n"
                    "      Do you want to overwrite it? (y/n)\n", stdout);
            } else {
                fputs("\n"
                    "Info: a file with the same NAME *but* with different SIZE already exists\n"
                    "      Do you want to overwrite it? (y/n)\n", stdout);
            }
            fflush(stdin);
            do { err = fgetc(stdin); } while(err == '\n');
            if((err != 'y') && (err != 'Y')) {
                fputs("  *SKIPPED*\n", stdout);
                return(0);
            } else fputs("  *OVERWRITING*\n", stdout);
        }
    }

    buff = alloca(FILESZ + 1);
    if(!buff) std_err();

        /* recovery enabled ONLY during the session to avoid possible problems */
        /* if a file with the same name already exists, the download will start from 0 */

    currsz = 0;
    while(currsz < size) {
        if(!proxy) {
            len = snprintf(buff, FILESZ, GET, uri, host, offset + currsz, size + offset);
        } else {
            len = snprintf(buff, FILESZ, GETPROXY, host, port, uri, host, offset + currsz, size + offset);
        }

        startsck(sd);

        length = 0;
        length2 = 0;
        realfile = 0;
        fsize = 0;

        err = send(sd, buff, len, 0);
        if(err < 0) std_err();

        while(1) {
            err = recv(sd, buff + length2,  FILESZ - length2, 0);
            if(err < 0) std_err();
            if(!err) break;
            length += err;

            if(realfile) {
                if(fwrite(buff, err, 1, fd) != 1) io_err();
                if(fsize) {
                    if(percentage) printf("%3u%%\r", (length + currsz) * 100 / fsize);
                } else fputc('.', stdout);
            } else {
                buff[length] = 0;
                stri = strstr(buff, "\r\n\r\n");
                if(stri) {

                    stri += 4;

                        /* check if file exist (HTTP/1.1 200 OK) */
                    if(memcmp(buff + 9, "20", 2)) {
                        printf("\nError: File not found on the server, server error %.3s\n",
                            buff + 9);
                        close(sd);
                        return(0);
                     }

                        /* filesize */
                        /* contlen probably is a bit useless here (because we already have */
                        /* the size) but I prefer to use it for more security */

                    contlen = strcasestr(buff, CONTLEN);
                    if(contlen) {
                        fsize = atol(contlen + (sizeof(CONTLEN) - 1)) + currsz;

                        fd = fopen(filename, "wb");
                        if(!fd) std_err();

                        printf("               %u\r", fsize);
                    } else {
                        fputs("               no recovery\r", stdout);
                        fd = fopen(filename, "wb");
                        if(!fd) std_err();
                    }

                        /* we write on the file the latest bytes of the string */
                    length2 = 0;
                    length += buff - stri;
                    if(length) if(fwrite(stri, length, 1, fd) != 1) io_err();
                    realfile = 1;
                } else length2 = length;
            }
            if(length >= size) break;
        }
        currsz += length;
        close(sd);
    }

    close(sd);

        /* building ZIP header */
    if(fd) {
        central_dir = (struct central_dir *)central;
        tmp = central_dir->offset;
        central_dir->offset = 0;
        if(fwrite(central, TOTCDIRSZ, 1, fd) != 1) io_err();
        central_dir->offset = tmp;

        tmp = PK3;
        if(fwrite(&tmp, 4, 1, fd) != 1) io_err();

        end_central_dir = (struct end_central_dir *)buff;
        end_central_dir->disk_num       = 0;
        end_central_dir->disk_num_start = 0;
        end_central_dir->tot_ent_disk   = 1;
        end_central_dir->tot_ent_dir    = 1;
        end_central_dir->dir_size       = TOTCDIRSZ;
        end_central_dir->offset         = size + 1;
        end_central_dir->comm_len       = 0;
        if(fwrite(buff, ECDIRSZ, 1, fd) != 1) io_err();

        fclose(fd);
    }

    totdl++;

        /* extra check, sometimes needed */
    if(extracheck) {
        if(!(lamepatch && (central_dir->extra_len == 0xd))) {
            fd = fopen(filename, "rb");
            if(!fd) std_err();

                /* check MZ */
            if(fread(&tmptest, 2, 1, fd) != 1) io_err();
            if(tmptest == 0x5a4d) {
                fputs("\nError: the downloaded file was from a bad auto-exctracting file and I'm not able to download it\n", stdout);
                exit(1);
            }

            err = fseek(fd, 28, SEEK_SET);
            if(err < 0) std_err();

                /* check extra field */
            if(fread(&tmptest, 2, 1, fd) != 1) io_err();
            if(central_dir->extra_len != tmptest) {
                fclose(fd);
                return(tmptest);
            }
            fclose(fd);
            fputs("- File seems correct\n", stdout);
        }
    }

    fputc('\n', stdout);
    return(0);
}



int show_zip(u_char *central) {
    u_char      *filename,
                tmpnull;
    u_short     extra,      /* patch */
                err;
    u_int      crc;

        /* initial check, very useful */
    central_dir = (struct central_dir *)central;
    if(central_dir->sign != PK2) return(-1);

        /* Q: why I don't check if the data I'm reading is going over the zip buffer? */
        /* A: I prefer to don't waste other CPU cicles because I check the PK2 sign */

    while(1) {
        if(central_dir->name_len > MAXNAMESZ) {
            printf("\nUhmmm this file seems to be corrupted or hacked, there is a filename long %hu bytes!!!\n",
                central_dir->name_len);
            return(-1);
        }
        filename = central + CDIRSZ;
        tmpnull = filename[central_dir->name_len];  // NULLify latest char (for visualization)
        filename[central_dir->name_len] = 0;

        printf("\n"
            "----------------------------------------------------------------\n"
            "Filename:     %s\n"
            "Zipped_size:  %-10u   File_size: %-10u   CRC: %08X\n"
            "File_offset:  %08x     Date:      %s\n",
            filename,
            central_dir->comp_size,
            central_dir->uncomp_size,
            central_dir->crc,
            central_dir->offset,
            maketime(central_dir->timedate));

        if(filecompare) {
            crc = calc_crc(filename);
            if(crc == 0xffffffff) {
                fputs("CRC-check:    *ERROR* local file doesn't exist\n", stdout);
            } else if(central_dir->crc != crc) {
                printf("CRC-check:    *ERROR* differents, local file CRC: %08x\n", crc);
            } else fputs("CRC-check:    *OK*\n", stdout);
        }

        filename[central_dir->name_len] = tmpnull;      // restore char

        if(dlarg) {
            if(strcasestr(filename, dlarg)) {

                    /* extra data patch! Seems to be caused by Unix zip... I dunno... */
                if(lamepatch) {
                    if(central_dir->extra_len == 0xd) extra = 0x15;
                        else extra = central_dir->extra_len;
                } else extra = central_dir->extra_len;

                err = http2file(
                    filename,
                    central_dir->name_len,
                    central_dir->offset,
                    PK1OFF + central_dir->comp_size + central_dir->name_len + extra,
                    central,
                    0);
                if(err) {
                    totdl--;
                    printf("\n\n"
                        "- The local ZIP header (extra length field) differs from the central (%hu -> %hu)\n"
                        "  This happens with some rare files, however it is not a problem, I TRY to redownload it correctly\n",
                        err, central_dir->extra_len);
                    http2file(
                        filename,
                        central_dir->name_len,
                        central_dir->offset,
                        PK1OFF + central_dir->comp_size + central_dir->name_len + err,
                        central,
                        1);
                }
            }
        }

        central += TOTCDIRSZ;

        central_dir = (struct central_dir *)central;
        if(central_dir->sign != PK2) break;
    }

    return(0);
}



u_int calc_crc(u_char *filename) {
    FILE    *fd;
    u_char  *buff;
    int     err;
    u_int  crc = 0xffffffff;

    fd = fopen(filename, "rb");
    if(!fd) return(0xffffffff);

    buff = malloc(BUFFSZ);
    if(!buff) std_err();

    while(!feof(fd)) {
        err = fread(buff, 1, BUFFSZ, fd);
        if(!err) break;
        crc = crc32(crc, buff, err);
    }

    fclose(fd);
    free(buff);
    return(~crc);
}



u_char *spc_conv(u_char *arg) {
    u_char          *ptr1,
                    *ptr2;
    static u_char   *buff;
    u_int          spc = 0,
                    len,
                    tmp,
                    i;

    ptr1 = arg;
    while(1) {
        ptr2 = strchr(ptr1, ' ');
        if(!ptr2) break;
        spc++;
        ptr1 = ptr2 + 1;
    }

    if(!spc) return(arg);

    len = strlen(arg) + (spc << 1) + 1;
    buff = malloc(len + 1);
    if(!buff) std_err();

    strcpy(buff, arg);
    ptr1 = arg;
    for(i = 1; i <= spc; i++) {
        ptr1 = strchr(ptr1, 0x20) + 1;
        tmp = ptr1 - arg;
        ptr2 = buff + tmp + (i << 1);
        memcpy(ptr2, ptr1, len - tmp);
        memcpy(ptr2 - 3, "%20", 3);
    }
    buff[len] = 0;
    return(buff);
}



void geturl(char *buff) {
    char        *tmp;

    if(!strncasecmp(buff, "ftp:", 4)) {
        fputs("\nError: FTP is not supported, the program supports only HTTP with resuming support\n", stdout);
        exit(1);
    }

    host = strstr(buff, "//");
    if(host) host += 2;
       else host = buff;
    uri = strchr(host, '/');
    if(uri) {
        *uri = 0;
        uri++;
    } else uri = "";
    tmp = strrchr(host, ':');
    if(tmp) {
        port = atoi(tmp + 1);
        *tmp = 0;
    }
}



u_int resolv(char *hostchar) {
    struct  hostent *hp;
    u_int  host_ip;

    host_ip = inet_addr(hostchar);
    if(host_ip == INADDR_NONE) {
        hp = gethostbyname(hostchar);
        if(!hp) {
            printf("\nError: Unable to resolve hostname (%s)\n", hostchar);
            exit(1);
        } else host_ip = *(u_int *)(hp->h_addr);
    }

    return(host_ip);
}



char *maketime(u_int timedate) {
    static  char    buff[23];
    u_short yr,
            mo,
            dy,
            hh,
            mm,
            ss,
            timez = timedate,
            datez = timedate >> 16;

    yr = ((datez >> 9) & 0x7f) + 1980;
    mo = ((datez >> 5) & 0x0f);
    dy = datez & 0x1f;

    hh = (timez >> 11) & 0x1f;
    mm = (timez >> 5) & 0x3f;
    ss = (timez & 0x1f) * 2;

    snprintf(buff, 22, "%02u:%02u:%02u %02u/%02u/%04u", hh, mm, ss, dy, mo, yr);
    return(buff);
}



void io_err(void) {
    fputs("\nError: I/O error during file writing\n", stdout);
    exit(1);
}



#ifndef WIN32
    void std_err(void) {
        perror("\nError");
        exit(1);
    }
#endif

